/*  cocodisk.c

	This program will display the directory of a Color Computer disk
	(RSDOS) and extract specified files under a Linux system. The program
	was developed on a Linux 2.0.0 with a HD 5.25 disk drive as /dev/fd1.
	Read the enclosed README for info on what to do before running this
	program.
	
	Bob van der Poel	bvdpoel@uniserve.com
	PO Box 57, Wynndel, BC, Canada, V0B 2N0
	or
	PO Box 355, Porthill, ID, USA, 83853
	
	
	Usage cocodisk [-D=disk_image_name] [-d] [-x] [-1] [filename...]  [-?]
	
	
	History:
	
	Version 1.0	January  1997
	Version 2.0	November 1999	Brian W. Hall <brihall@bigfoot.com>
	
	V2 has been modified to operate on DSK images only, since these are
	widely available for emulators. It should be possible to create a DSK
	image from Linux, by using the setfdprm command as described in the V1
	documentation, and then issuing this command:
		dd if=/dev/fd1 of=COCO1.DSK
	
	For a Linux CoCo emulator, use xmess:
		http://xmame.retrogames.com/
		
*/

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdarg.h>

struct
{
	u_char filename[8];
	u_char extension[3];
	u_char filetype;
	u_char asciiflag;
	u_char firstgran;
	u_char bytes[2];
	u_char reserved[16];
} coco_dir_ent[72];	/* The entire coco directory-track 17, sectors 3-11 */

u_char grans[72];	/* the file allocation table-track 17, sector 2     */

FILE *CoCodisk;		/* input file when reading from CoCo disk image	    */
int outpath;		/* output file when writing to Linux		    */

void	bstrip(char *),
	copy(int, int, int),
	do_short_dir(void),
	extract(char *, char *),
	extract_all(void),
	extract_one(char *),
	normalize(char *, char *, char *),
	terminate(char *, ...),
	usage(void),
 	do_directory(void);


enum{FILES=0, SINGLE_DIR, FULL_DIR, GET_ALL};

/*	======================================================================
	MAIN
*/

int
main(int argc, char **argv)
{
	char *coco_disk="D1.DSK";
	int command=FILES;
	int t;
	
	while((t=getopt(argc, argv, "D:d1x?"))!=-1)
	{
		switch(t)
		{
			case 'D':
				coco_disk=optarg;
				break;
				
			
			case '1':
				command=SINGLE_DIR;
				break;
				
			case 'd':
				command=FULL_DIR;
				break;
				
			case 'x':
				command=GET_ALL;
				break;
				
			case '?':
			default:
				usage();
				break;
		}
	}
	
	if(optind==argc && command==FILES)
	{
		puts("\nNo commands or filenames given...I've gotta have something to do!");
		 usage();
	}
	
	if(optind<argc && command!=FILES)
	{
		puts("\nPlease enter a command OR filename(s). I get confused by both!");
		usage();
	}
	
	if ((CoCodisk = fopen (coco_disk, "r")) == NULL)
	{
		printf ("Could not open disk file name %s\n", coco_disk);
		exit (1);
	}
	
	// grab the granual allocation table, track 17, sector 2
	
	fseek (CoCodisk, ((17*18)+1)*256, SEEK_SET);
	if (fread (&grans, sizeof(grans), 1, CoCodisk) != 1)
		terminate("Error reading FAT, %s", strerror(errno));
	
	// the directory starts at track 17, sector 3
	// read the whole thing into memory
	
	fseek( CoCodisk, ((17*18)+2)*256, SEEK_SET);
	if (fread (&coco_dir_ent, sizeof(coco_dir_ent), 1, CoCodisk) != 1)
		terminate("Error reading directory, %s", strerror(errno));
	
	switch(command)
	{
		case FULL_DIR:		do_directory();		break;
		case SINGLE_DIR:	do_short_dir();		break;
		case GET_ALL:		extract_all();		break;
		
		case FILES:
			for(t=optind; t<argc; extract_one(argv[t++]));
			break;
			
		default:
			terminate("Unknown command! Impossible.");
	}
	exit(0);
}



/*	======================================================================
	Display a full directory
*/

void
do_directory(void)
{
	int t, col, g;
	char fname[10];		// these needed since there is no
	char fext[5];		// 0 terms on the filename/extension
	char aflag;
	int fsize, gcount, gused;
	
	fname[8]='\0';		// we'll be copying non-terminated data to these
	fext[3]='\0';
	
	for(col=0,t=0, gused=0; t<72; t++)
	{
		if(coco_dir_ent[t].filename[0]=='\0') continue;	// deleted entry
		if(coco_dir_ent[t].filename[0]==0xff) break;		// end of directory
		
		// grab the filename & extension
		
		memmove(fname, coco_dir_ent[t].filename, 8);
		memmove(fext, coco_dir_ent[t].extension, 3);
		
		aflag=(coco_dir_ent[t].asciiflag=0xff) ? 'B' : 'A';
		
		fsize=(coco_dir_ent[t].bytes[0]*256)+coco_dir_ent[t].bytes[1];
		gcount=0;
		for(g=coco_dir_ent[t].firstgran; ;)
		{
			if(grans[g]<=0x43)
			{
				g=grans[g];
				fsize+=(256*9);
				gcount++;
				continue;
			}
			
			if(grans[g]>=0xc0 && grans[g]<=0xc9)
			{
				g=grans[g] & 0x1f;	// keep bits 0-5
				fsize+=(g-1)*256;	// last sect. already added in
				gcount++;
				break;
			}
			
			terminate("Error in file allocation table, byte %x==%x, file %s %s\n",
				g, grans[g], fname, fext);
		}
		
		printf("%8s %3s  %d  %c %5d/%2d    ", fname, fext,
			coco_dir_ent[t].filetype, aflag, fsize, gcount);
		gused+=gcount;
		if(++col==2) printf("\n"), col=0;
	}
	if(col!=0) printf("\n");
	
	printf("Grans allocated   %d\n", gused);
	for(t=0, g=0; g<68; g++) if(grans[g]==0xff) t++;
	printf("Grans free in FAT %d\n", t);
}



/*	======================================================================
	Display filenames in a single column. This is useful to send to a file which
	can be edited and then used as input for a subsequent extraction of selected files.
	I'm too lazy to write a wildcarding routine for cocodisk <g>
*/

void
do_short_dir(void)
{
	int t;
	char fname[20];		// normalized name
	
	for(t=0; t<72; t++)
	{
		if(coco_dir_ent[t].filename[0]=='\0') continue;	// deleted entry
		if(coco_dir_ent[t].filename[0]==0xff) break;		// end of directory
		
		// grab the filename & extension
		
		normalize(fname, coco_dir_ent[t].filename, coco_dir_ent[t].extension);
		printf("%s\n", fname);
	}
}


/*	======================================================================
	Extract all the files on this disk.
*/

void
extract_all()
{
	int t;
	
	for(t=0; t<72; t++)
	{
		if(coco_dir_ent[t].filename[0]=='\0') continue;	// deleted entry
		if(coco_dir_ent[t].filename[0]==0xff) break;		// end of directory
		
		extract(coco_dir_ent[t].filename, coco_dir_ent[t].extension);
	}
}


/*	======================================================================
	Extract a specifed file. The filename is in Unix format, we need to convert to
	RSDOS format and pass to extract().
*/


void
extract_one(char *s)
{
	char fname[10];
	char ext[5];
	char *p;
	
	memset(fname, ' ', sizeof(fname));
	memset(ext, ' ', sizeof(ext));
	
	p=index(s, '.');
	if(p==NULL) p=s+strlen(s);
	if(p-s>8) terminate("RSDOS filename %s too long", s);
	memmove(fname, s, p-s);
	if(*p)
	{
		p++;
		if(strlen(p)>3) terminate("RSDOS filename extension in %s too long", s);
		strcpy(ext, p);
	}
	
	extract(fname, ext);
}


/*	======================================================================
	Extract a file
	
	name -- 8 char name, blank filled
	ext  -- 3 char ext,     "    "   
*/

void
extract(char *name, char *ext)
{
	int f, g;
	char fname[20];
	int path;
	
	for(f=0; f<68; f++)
	{
		if(coco_dir_ent[f].filename[0]==0xff)
			terminate("Filename %s%s not found", name, ext);
		if((memcmp(coco_dir_ent[f].filename, name, 8)==0)
			&& (memcmp(coco_dir_ent[f].extension, ext, 3)==0))  // found name
			break;
	}
	
	normalize(fname, name, ext);
	path=open(fname, O_CREAT|O_EXCL|O_WRONLY, 0666);
	if(path<=0) terminate("Can't open %s for writing, %s", fname, strerror(errno));
	
	printf("Copying %s\n", fname);
	for(g=coco_dir_ent[f].firstgran; ; )
	{
		if(grans[g]<=0x43)
		{
			copy(path, g, 9*256);
			g=grans[g];
			continue;
		}
		if(grans[g]>=0xc0 && grans[g]<=0xc9)
		{
			copy(path, g, (((grans[g] & 0x1f)-1)*256)+
				(coco_dir_ent[f].bytes[0]*256)+coco_dir_ent[f].bytes[1]); 
			break;
		}
	}
	close(path);
}

/*	======================================================================
	copy from RSDOS to linux
	
	path	- the output path for the linux file
	gran	- the granule to write
	sz	- number of bytes from this gran to copy
*/

void
copy(int path, int gran, int sz)
{
	char buff[9*256];
	
	if(gran>32) gran+=2;

	fseek (CoCodisk, gran*9*256, SEEK_SET);
	if (fread (&buff, sz, 1, CoCodisk) != 1)
		terminate("Read error, %s", strerror(errno));
	if (write(path, buff, sz)!=sz)
		terminate("Write error, %s", strerror(errno));
}


/*	======================================================================
	Strip off trailing spaces from a string
*/

void
bstrip(char *str)
{
	char *p;
	
	p=str;
	while(*p) p++;	// move to end of string
	p--;		// last char in string
	while(*p==' ' && p>=str)
	{
		*p='\0';
		p--;
	}
}


/*	======================================================================
	convert the 8/3 RSDOS directory name to a unix-like string
	
	buff	- where the new names goes...make sure it's big enuf
	name	- original main part of the name, 8 chars, not 0 terminated
	ext	-  "       extension...
*/

void
normalize(char *buff, char *name, char *ext)
{
	char fname[10], fext[5];
	
	memmove(fname, name, 8);
	fname[8]='\0';
	bstrip(fname);
	memmove(fext, ext, 3);
	fext[3]='\0';
	bstrip(fext);
	strcpy(buff, fname);
	if(*fext)
	{
		strcat(buff, ".");
		strcat(buff, fext);
	}
}

/*	======================================================================
	Do the usage message and exit
*/

void
usage(void)
{

	static char *msg[]=
	{
		"",
		"cocodisk - (c) 1997 Bob van der Poel Software",
		"v2 modifications by Brian Hall <brihall@bigfoot.com>",
		"",
		"Extracts files/directories from "
		"Tandy Color Computer disk images",
		"Usage: [opts] [files] [opts]",
		"Options:",
		" -D image	"
		"specify disk image for read (default D1.DSK)",
		" -d		display full directory of disk",
		" -1		display short form directory",
		" -x		extract all files and create them in current dir",
		" files		extract specified file(s) to current dir",
		"",
		" Please read the documentation which came with this package!",
		""
	};
	
	int t;
	
	for(t=0; t<sizeof(msg)/sizeof(msg[0]); puts(msg[t++]));
	
	exit(1);
}


/*	======================================================================
	Fatal errors come here. Display a message and exit.
*/


void
terminate(char *s, ...)
{
	va_list ap;
	
	va_start(ap, s);
	vprintf(s, ap);
	va_end(ap);
	
	exit(1);
}
